import  os
import glob
import numpy as np
from scipy.io.wavfile import read
import tensorflow as tf
from  createSimpleSegmentationData1rec import get1wav_targSegmentation_sig
#from getChalengeData import read_wav_files_and_labels
import helper_code as hc
from getOnlyChallengeDataLabels import getLabels

dataDir = '/home/rui/ensino/adne2022/ecgSounds/challenge2022/training_data/'
#dataDir = '/home/rui/research/challenge2022/official/trainData_1.03/training_data/'
#dataDir = '/home/rui/research/challenge2022/official/challengeData_103/training_data/'

# stft configurations
frameL = 128
frameStepFactor = 2#4
frameStep = frameL // frameStepFactor
maxFreq = frameL//2 + 1

def compSTFT(array):
    spec = tf.signal.stft(array,frameL, frameStep)
    spec = spec[:,:maxFreq]
    spec = tf.math.abs(spec)
    return spec

subsamplingFactor = 8 # segmentation signal

# python function to apply in the tf.data.dataset reads exactly the all signal and target
def getMultitaskItemsComplete(wavrecname):
    # recname endswith '.wav'
    wavrecname = wavrecname.decode()
    
    murmurLabel, outcomeLabel = getLabels(wavrecname)
    sigSegment, targ = get1wav_targSegmentation_sig(wavrecname[:-4])#, '')
    return sigSegment, murmurLabel, outcomeLabel, targ, [wavrecname]

# tf function
# it uses a random segment (fraction) of  the  signal length

#frac = 2/3

def randomCutSignalTarg( sigSegment, murmurLabel, outcomeLabel, targ, wavrecname ):
    
    l = tf.shape(sigSegment)[0] # sigSegment length

    frac = tf.random.uniform([], minval= 0.5, maxval=1.0, dtype=tf.float32)
    
    # targ and sigSegment not subsampled
    segmented = targ>0
    aux1 = tf.argmax(segmented)+100-tf.cast(frac*tf.cast(l,tf.float32),tf.int64)
    aux2 = tf.cast(l,tf.int64)-tf.argmax(segmented[::-1])-100
    
    nFrameStepsSignal = tf.cast(l / frameStep, tf.int32)
    startLimitEnd = tf.cast((1-frac)*tf.cast(nFrameStepsSignal*frameStep,tf.float32), tf.int64)
    startLimitEnd = tf.minimum(startLimitEnd, aux2)
    startLimitBegin = tf.maximum(tf.cast(0,tf.int64),aux1)                       
    #start = tf.random.uniform([], minval=startLimitBegin, maxval=startLimitEnd, dtype=tf.int64)
    start = tf.cond(startLimitEnd > startLimitBegin,
                    lambda : tf.random.uniform([], minval=startLimitBegin,
                                               maxval=startLimitEnd, dtype=tf.int64),
                    lambda: tf.cast(0,tf.int64))
    newl = tf.cast(frac * tf.cast(nFrameStepsSignal,tf.float32),tf.int32)*frameStep
    sigSegment = sigSegment[start:start+tf.cast(newl,tf.int64)]
    targ = targ[start:start+tf.cast(newl,tf.int64):subsamplingFactor]
    return sigSegment, murmurLabel, outcomeLabel, targ, [wavrecname], start

def compSTFT_allsig_segment( sigSegment,
                             murmurLabel, outcomeLabel, segmentTarg, wavrecname,start):
    segmentSpec = compSTFT(sigSegment)
    return segmentSpec,  murmurLabel, outcomeLabel, segmentTarg#, wavrecname,start

def separateTuples( segmentSpec, murmurLabel, outcomeLabel, segmentTarg):#, wavrecname):
    return (segmentSpec,(murmurLabel, outcomeLabel, segmentTarg))#,  [wavrecname], start))

def getDataset(wavFileList, batch_size):
    dataset = tf.data.Dataset.from_tensor_slices(wavFileList)
    dataset = dataset.map(
        lambda item: tuple(tf.numpy_function( getMultitaskItemsComplete, [item],
                                             [tf.float32,tf.float32,tf.float32,tf.float32, tf.string])))
    dataset = dataset.map(randomCutSignalTarg)
    specdataset = dataset.map(compSTFT_allsig_segment)
    bdataset = specdataset.padded_batch(batch_size,
                                         padded_shapes=([None, maxFreq],[],[],[None]))             
    bdataset = bdataset.map(separateTuples)
    return bdataset



# def getAllWavFiles(data_folder):
#     patient_files = hc.find_patient_files(data_folder)
#     allwavfiles = []
#     for patient in patient_files:
#         patient_header_data = hc.load_patient_data(patient)
#         # follow 'load_recordings'
#         num_locations = hc.get_num_locations(patient_header_data)
#         recording_information = patient_header_data.split('\n')[1:num_locations+1]
#         for i in range(num_locations):
#             entries = recording_information[i].split(' ')
#             recording_file = entries[2]
#             filename = os.path.join(data_folder, recording_file)
#             allwavfiles.append(filename)
#     return allwavfiles

#allwavfiles = getAllWavFiles(dataDir)
#batchSise = 4
#experdataset = getDataset(allwavfiles, batchSise)
#it = iter(experdataset)
#ex = next(it)
