import sys
import glob
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
import pdb
from evaluate_model import  cost_algorithm, cost_expert, cost_treatment, cost_error




############## custom losses & metrics ##########


####### Segmentation

scceLogits =tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True,
        reduction=tf.keras.losses.Reduction.NONE)

def accu0(y_true, y_pred):
    notzero = tf.cast(tf.greater(tf.cast(y_true,tf.int64), tf.constant(0,dtype=tf.int64)),tf.int64)
    aux =notzero*tf.cast(tf.equal(tf.cast(y_true,tf.int64),tf.argmax(y_pred,axis=-1)),tf.int64)
    numNonZero = tf.reduce_sum(notzero)
    res = tf.reduce_sum(aux)/numNonZero
    return res



def xEntrop0(y_true, y_pred_logits):
    tf.assert_equal(tf.shape(y_true)[1],tf.shape(y_pred_logits)[1])
    notzero = tf.cast(tf.greater(y_true, tf.constant(0.0)),tf.float32)
    numNonZero = tf.reduce_sum(notzero,axis=1)
    cross = scceLogits(y_true, y_pred_logits)
    res = tf.reduce_sum(notzero * cross, axis=1)/numNonZero
    #tf.debugging.check_numerics(res,"xentropy0")
    return res

###### Murmurios


## Loss adaptada à classificação dos murmúrios:

numClassMurmur = 3


scceNoLogits =tf.keras.losses.SparseCategoricalCrossentropy(from_logits=False,
        reduction=tf.keras.losses.Reduction.NONE)




def modified_cost_s_murmur(y_label, y_pred):
    y_true = tf.one_hot(tf.cast(y_label,tf.int32), numClassMurmur)
    y_true = tf.reshape(y_true,(tf.shape(y_true)[0],tf.shape(y_true)[-1]))

    l=5*(y_pred[:,1]*y_true[:,0]+y_pred[:,2]*y_true[:,0])\
     +3*(y_pred[:,0]*y_true[:,1]+y_pred[:,2]*y_true[:,1])\
     +(y_pred[:,0]*y_true[:,2]+y_pred[:,1]*y_true[:,2])

    return l + tf.keras.losses.SparseCategoricalCrossentropy()(y_label, y_pred)



#  metric with absolute classification
class true_s_murmur(tf.keras.metrics.Metric):
    # classes = ['Positive', 'Unknown', 'Absent']

    def __init__(self, name='true_s_murmur', **kwargs):
        super(true_s_murmur, self).__init__(name=name, **kwargs)
        self.correctPositives = self.add_weight(name='cP', initializer='zeros')
        self.correctUnknowns = self.add_weight(name='cU', initializer='zeros')
        self.correctAbsents = self.add_weight(name='cU', initializer='zeros')
        self.Pos = self.add_weight(name='wp', initializer='zeros')
        self.Unk = self.add_weight(name='wu', initializer='zeros')
        self.Abs = self.add_weight(name='wa', initializer='zeros')

    def update_state(self, y_true, y_pred, sample_weight=None): 
        y_pred = tf.argmax(y_pred, axis=-1)
        y_true = tf.cast(y_true, tf.int64)
        y_true = tf.reshape(y_true, (tf.shape(y_true)[0],))
        

        correctPos = tf.cast(tf.equal(y_true,0), tf.float32)*tf.cast(tf.equal(y_pred,0), tf.float32)
        correctUnk = tf.cast(tf.equal(y_true,1), tf.float32)*tf.cast(tf.equal(y_pred,1), tf.float32)
        correctAbs = tf.cast(tf.equal(y_true,2), tf.float32)*tf.cast(tf.equal(y_pred,2), tf.float32)        
  
        self.correctPositives.assign_add(tf.reduce_sum(correctPos))
        self.correctUnknowns.assign_add(tf.reduce_sum(correctUnk))
        self.correctAbsents.assign_add(tf.reduce_sum(correctAbs))
                                       
        # true positive wrongly classified
        wrongPosBatch = tf.cast(tf.equal(y_true,0), tf.float32)*tf.cast(tf.not_equal(y_pred,0), tf.float32)
        
        self.Pos.assign_add(tf.reduce_sum(wrongPosBatch+ correctPos))

        # true unknown wrongly classified
        wrongUnkBatch = tf.cast(tf.equal(y_true,1), tf.float32)*tf.cast(tf.not_equal(y_pred,1), tf.float32)
        self.Unk.assign_add(tf.reduce_sum(wrongUnkBatch + correctUnk))

        # true absent wrongly classified
        wrongAbsBatch = tf.cast(tf.equal(y_true,2), tf.float32)*tf.cast(tf.not_equal(y_pred,2), tf.float32)
        self.Abs.assign_add(tf.reduce_sum(wrongAbsBatch + correctAbs))

     
    def result(self):                                      
        l = (5*self.correctPositives + 3*self.correctUnknowns + self.correctAbsents)/(5.0*self.Pos+3*self.Unk+self.Abs)
        return l
        
    def reset_states(self):
        self.correctPositives.assign(0)
        self.correctUnknowns.assign(0)
        self.correctAbsents.assign(0)        
        self.Pos.assign(0)
        self.Unk.assign(0)
        self.Abs.assign(0)
    


########################## outcome
        
# ## Loss adpatada ao custo da classificação Normal/Abnormal



# fórmula simplificada que 'contém o essencial'
def modified_c_outcome(y_true, y_pred):
    # y_true=0 for class Ab=0 y_true=1 for class Normal=1
    # y_pred = prob of being equal to class Normal=1
    #
    y_pred = tf.reshape(y_pred,(-1,))
    y_true = tf.reshape(y_true,(-1,))
    
    cout = (1-y_pred)*y_true+4**y_pred*(1-y_true)
    
    return cout

def c_outcome_crossentropy(y_true, y_pred):
    factorOutcome = 0.00001
    crossentropy = tf.keras.losses.BinaryCrossentropy()(y_true, y_pred)
    return  crossentropy + factorOutcome*c_outcome(y_true, y_pred)


def modified_c_outcome_crossentropy(y_true, y_pred):
    factorOutcome =1.0
    crossentropy = tf.keras.losses.BinaryCrossentropy()(y_true, y_pred)
    return  crossentropy + factorOutcome*modified_c_outcome(y_true, y_pred)

## metric_c_outcome

class metric_c_outcome(tf.keras.metrics.Metric):
    # classes = ['Abnormal', 'Normal'] 'Abnormal'->y_true=0,  'Normal'->y_true=1

    def __init__(self, name='metric_c_outcome', **kwargs):
        super(metric_c_outcome, self).__init__(name=name, **kwargs)
        self.pred_abnormal = self.add_weight(name='pred_abnormal', initializer='zeros')
        self.false_normal = self.add_weight(name='false_normal', initializer='zeros')
        self.true_abnormal = self.add_weight(name='true_abnormal', initializer='zeros')
        self.total = self.add_weight(name='total', initializer='zeros')

    

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_pred = tf.reshape(y_pred,(-1,))
        y_pred = tf.round(y_pred)
        y_true = tf.reshape(y_true,(-1,))
        
        pred_abnormal_batch = tf.reduce_sum(1-y_pred)
        self.pred_abnormal.assign_add(pred_abnormal_batch)
            
        false_normal_batch = tf.reduce_sum((1-y_true)*y_pred)
        self.false_normal.assign_add(false_normal_batch)
        
        true_abnormal_batch = tf.reduce_sum( (1-y_true)*(1-y_pred))
        self.true_abnormal.assign_add(true_abnormal_batch)
        batch_size = tf.cast(tf.shape(y_true)[0],tf.float32)
        self.total.assign_add(batch_size)
  

    def result(self):
        l =   cost_algorithm( self.total)\
            + cost_expert(self.pred_abnormal, self.total) \
            + cost_treatment(self.true_abnormal) \
            + cost_error(self.false_normal)
        return l/self.total
        
    def reset_state(self):
        self.pred_abnormal.assign(0)
        self.false_normal.assign(0)
        self.true_abnormal.assign(0)
        self.total.assign(0)

        
#############################
### define model ###
##############################





class PositionEmbedding(layers.Layer):
    def __init__(self, maxlen, embed_dim):
        super(PositionEmbedding, self).__init__()
        self.pos_emb = layers.Embedding(input_dim=maxlen, output_dim=embed_dim)

    def call(self, x):
        maxlen = tf.shape(x)[1]
        batchS = tf.shape(x)[0]
        positions = tf.range(start=0, limit=maxlen, delta=1)
        positions = self.pos_emb(positions)
        positions = tf.expand_dims(positions, axis = 0)
        positions = tf.repeat(positions, batchS, axis=0)
        return x + positions


class TransformerBlock(layers.Layer):
    def __init__(self, embed_dim, num_heads, ff_dim, rate=0.1):
        super(TransformerBlock, self).__init__()
        self.att = layers.MultiHeadAttention(num_heads=num_heads, key_dim=embed_dim)
        self.ffn = keras.Sequential(
            [layers.Dense(ff_dim, activation="relu"), layers.Dense(embed_dim),]
        )
        self.layernorm1 = layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = layers.LayerNormalization(epsilon=1e-6)
        self.dropout1 = layers.Dropout(rate)
        self.dropout2 = layers.Dropout(rate)

    def call(self, inputs, training):
        attn_output = self.att(inputs, inputs)
        attn_output = self.dropout1(attn_output, training=training)
        out1 = self.layernorm1(inputs + attn_output)
        ffn_output = self.ffn(out1)
        ffn_output = self.dropout2(ffn_output, training=training)
        return self.layernorm2(out1 + ffn_output)


class TransformerBlockSpatialDrop(layers.Layer):
    def __init__(self, embed_dim, num_heads, ff_dim, rate=0.1):
        super(TransformerBlockSpatialDrop, self).__init__()
        self.att = layers.MultiHeadAttention(num_heads=num_heads, key_dim=embed_dim)
        self.ffn = keras.Sequential(
            [layers.Dense(ff_dim, activation="relu"), layers.Dense(embed_dim),]
        )
        self.layernorm1 = layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = layers.LayerNormalization(epsilon=1e-6)
        self.dropout1 = layers.SpatialDropout1D(rate)
        self.dropout2 = layers.SpatialDropout1D(rate)

    def call(self, inputs, training):
        attn_output = self.att(inputs, inputs)
        attn_output = self.dropout1(attn_output, training=training)
        out1 = self.layernorm1(inputs + attn_output)
        ffn_output = self.ffn(out1)
        ffn_output = self.dropout2(ffn_output, training=training)
        return self.layernorm2(out1 + ffn_output)

    
###################### convolution model ######################


# define convBlock as a subclass in order to set training = False in the batch normalization layer
# see https://keras.io/api/models/model/
class convBlock(tf.keras.Model):

    def __init__(self, nfilts, ksize=3, strides=2):
        super().__init__()
        self.conv = layers.Conv1D(nfilts, ksize, strides=1)
        self.BN = layers.BatchNormalization()
        self.activation = layers.Activation("relu")
        self.pool = layers.MaxPooling1D(ksize, strides=strides)

    def call(self, inputs, training=False):
        x = self.conv(inputs)
        x = self.BN(x, training=training)
        x = self.activation(x)
        x = self.pool(x)
        return x
         

        
# # stft configurations
# frameL = 128
# frameStepFactor = 4
# frameStep = frameL // frameStepFactor
# maxFreq = frameL//2 + 1

#subsamplingFactor = 8 # (same as in createMultitaskmodel.py)



class commonModel(tf.keras.Model):
    """3 tasks common block"""

    def __init__(self):
        super().__init__()
        
        ###### conv blocks
        nfilt1 = 32
        self.convBlock1 = convBlock(nfilt1, 5, 3)
        self.convBlock1._name = 'conv1'
        nfilt2 = 64
        self.convBlock2 = convBlock(nfilt2, 5, 3)
        self.convBlock2._name = 'conv2'
        nfilt3 = 128
        self.convBlock3 = convBlock(nfilt3, 3, 3)
        self.convBlock3._name = 'conv3'
    
        ## transformer blocks
        embed_dim = 64
        ff_dim = 128
        num_heads = 8

        
        # common(segmentation and classification)  transformers
        self.transformer1 = TransformerBlockSpatialDrop(embed_dim=embed_dim, ff_dim=ff_dim, num_heads=num_heads,
                                                   rate=0.1)
        self.transformer1._name='common_transformer1'
        self.transformer2 = TransformerBlockSpatialDrop(embed_dim=embed_dim, ff_dim=ff_dim, num_heads=num_heads,
                                               rate=0.1)
        self.transformer2._name='common_transformer2'
        self.transformer3 = TransformerBlockSpatialDrop(embed_dim=embed_dim, ff_dim=ff_dim, num_heads=num_heads,
                                               rate=0.1)
        self.transformer3._name='common_transformer3'
        
        spatialDropRate = 0.1
        self.spatialdrop1 = layers.SpatialDropout1D(spatialDropRate)
        self.spatialdrop2 = layers.SpatialDropout1D(spatialDropRate)
        self.spatialdrop3 = layers.SpatialDropout1D(spatialDropRate)
        
        self.reEncode = layers.TimeDistributed(layers.Dense(embed_dim))
        self.positionEmbedding = PositionEmbedding(1000,embed_dim) 
        
    def call(self, inputs, training=False):
        
        xs =  self.convBlock1(inputs, training=training)
        xs =  self.spatialdrop1(xs, training=training)
        xs =  self.convBlock2(xs, training=training)
        xs =  self.spatialdrop2(xs, training=training)
        xs =  self.convBlock3(xs, training=training)
        xs =  self.spatialdrop3(xs, training=training)
        
        xsNdim = self.reEncode(xs)
        txs = self.positionEmbedding(xsNdim)
        txs = self.transformer1(txs, training=training)
        txs = self.transformer2(txs, training=training)
        txs = self.transformer3(txs, training=training)
        return txs
        
        



def get_model(training,frameL, frameStepFactor, subsamplingFactor = 8, weights_file=None):
    
    frameStep = frameL // frameStepFactor
    maxFreq = frameL//2 + 1

    
    #####   3 tasks common block ###

    cModel = commonModel()

    inputs = keras.Input(shape=(None, maxFreq))

    txs = cModel(inputs, training=training)#False)
   

    ######### end of the common block
    
    ##################### segmentation block
    subsamplingFactor = 8
    numOutputs = 5
    l =9
    
    outx =layers.TimeDistributed(layers.Dense(l*numOutputs))(txs)
    outx = layers.Reshape((-1,numOutputs))(outx)
    
    convFactor = 27//l 
    upsamplingFactor = convFactor*(frameStep +6)//subsamplingFactor
    xupsample = layers.UpSampling1D(upsamplingFactor)(outx)
    shout_factor = frameStep
    aux = tf.shape(inputs)[1]
    shout = (aux + frameStepFactor -1)*shout_factor//subsamplingFactor
    segOutputs = xupsample[:,:shout]
    IdLayer = tf.keras.layers.Lambda(lambda x: x, name='segment')
    segOutputs = IdLayer(segOutputs)

    ################### Separated  Murmur and outcome classification blocks
    
    numClassesMurmur = 3
    murmurOutcomeEmbedingDim = 32
    murmurOutcomeDense_dim = 64
    num_heads_murmurOutcome = 8
    
    # murmur

    txcm =layers.TimeDistributed(layers.Dense(murmurOutcomeEmbedingDim))(txs)
    txcm = layers.SpatialDropout1D(0.5)(txcm)
    
    transformer3Murmur = TransformerBlockSpatialDrop(embed_dim=murmurOutcomeEmbedingDim,
                                                     num_heads=num_heads_murmurOutcome,
                                                     ff_dim=murmurOutcomeDense_dim, rate=0.5)
    transformer3Murmur._name='murmur_transformer'
    
    txcm = transformer3Murmur(txcm)
    txcm_0 = txcm[:,0]
    txcm_0 = layers.Dropout(0.5)(txcm_0)
    hidMurmur = layers.Dense(murmurOutcomeDense_dim) 
    txmurmur = hidMurmur(txcm_0)
    txmurmur = layers.Dropout(0.5)(txmurmur)
    classMurmur = layers.Dense(numClassesMurmur, activation='softmax')(txmurmur)
    classMurmur = layers.Flatten(name='murmur')(classMurmur)


    #Outcome
    
    #txconcat2 = layers.Concatenate(axis=-1)([txseg, txs, txcm])
    
    txco =layers.TimeDistributed(layers.Dense(murmurOutcomeEmbedingDim))(txs)
    txco = layers.SpatialDropout1D(0.75)(txco)
    transformer3Outcome = TransformerBlockSpatialDrop(embed_dim=murmurOutcomeEmbedingDim, num_heads=8,
                                                      ff_dim=murmurOutcomeDense_dim, rate=0.5)
    transformer3Outcome._name='outcome_transformer'
    
    txco = transformer3Outcome(txco)
    txco_0 = txco[:,0]
    txco_0 = layers.Dropout(0.75)(txco_0)

    
    hidOutcome = layers.Dense(murmurOutcomeDense_dim)
    txOutcome = hidOutcome(txco_0)
    txOutcome = layers.Dropout(0.75)(txOutcome)    
    classOutcome = layers.Dense(1, activation='sigmoid')(txOutcome)    
    classOutcome = layers.Flatten(name='outcome')(classOutcome)

    
    classificationModel = keras.Model(inputs, outputs=[classMurmur, classOutcome])
    multitaskModel = keras.Model(inputs=inputs,
                                 outputs = [classMurmur, classOutcome,segOutputs])
    
    
    lr = 1e-4
    opt = keras.optimizers.Adam(learning_rate=lr, clipvalue = 0.1)
    losses={'murmur':modified_cost_s_murmur, 'outcome': modified_c_outcome_crossentropy, 'segment':xEntrop0}
    weight_losses={'murmur': 1.0, 'outcome': 5, 'segment': 1}

    true_murmur = true_s_murmur()
    multitaskModel.compile(optimizer=opt,  loss=losses,
                           metrics = {'murmur': true_murmur,#'sparse_categorical_accuracy',
                                      'outcome': [metric_c_outcome(),'binary_accuracy'],
                                      'segment':accu0},
                           loss_weights=weight_losses)


    print('multitask model weights: ',weights_file)
    if not weights_file is None:
        print('loading multitask model weights: ',weights_file)
        multitaskModel.load_weights(weights_file)


    return multitaskModel, classificationModel, cModel #, segmentationModel, murmurModel, outcomeModel
