#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################
import os
#os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

import glob

import pdb

from helper_code import *
import numpy as np, scipy as sp, scipy.stats, os, sys, joblib
from sklearn.impute import SimpleImputer
from sklearn.ensemble import RandomForestClassifier

# our teams imports
import tensorflow as tf
#import class3CustomLossMultitaskModel as mtM
#import createTrainingDataset as cTD
from createCutMultitaskDataset import  getDataset
from newmultitaskModel2offi import get_model

################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################


pretrained_weights_file =  'commonOffi2_weights.h5'
trained_weights_file = 'trained_model.h5'

# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):
    # Find data files.
    if verbose >= 1:
        print('Finding data files...')


    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)
    print('model folder name is ',model_folder)


    # stft configurations (same as in 'createMultitaskDataset.py')
    frameL = 128
    frameStepFactor = 2#4
    frameStep = frameL // frameStepFactor
    maxFreq = frameL//2 + 1

    weightsfilepath = pretrained_weights_file
    multitask_model, classificationModel, cModel = \
        get_model(False, frameL, frameStepFactor)

    ## load pre-trained weighs,  freeze pre-trained layers and
    # only train the other layers
    cModel.load_weights('commonOffi2_weights.h5')
    cModel.trainable = False


    ## the following code is for submission
    #
    #####################################
    allwavfiles = []
    for patient in patient_files:
        patient_header_data = load_patient_data(patient)
        # follow 'load_recordings'
        num_locations = get_num_locations(patient_header_data)
        recording_information = patient_header_data.split('\n')[1:num_locations+1]
        for i in range(num_locations):
            entries = recording_information[i].split(' ')
            recording_file = entries[2]
            filename = os.path.join(data_folder, recording_file)
            simplename = os.path.basename(filename) 
            if not simplename.startswith('50782_MV_1'):
                allwavfiles.append(filename)


    #allwavfiles = [os.path.join(data_folder, f) for f in allwavfiles]
    batch_size = 64
    

    trainDataset = getDataset(allwavfiles, batch_size)
    ## debug
    # pdb.set_trace()
    #it = iter(trainDataset)
    #ex = next(it)
    
    numEpochs = 300
    multitask_model.fit(trainDataset, epochs=numEpochs)
    
    #########################################
    # ## this is for 'home use'
    # valFolder = 2
    # train_wav_files, val_wav_files = getTrainWavValFiles( valFolder)
    # train_wav_files = [data_folder + f for f in train_wav_files] # corrigir !!!
    # val_wav_files = [data_folder + f for f in val_wav_files] # corrigir !!!
    # trainDataset = getDataset(train_wav_files, batch_size)
    # valDataset = getDataset(val_wav_files, batch_size)

    ## debug
    #it = iter(trainDataset)
    #ex = next(it)   
    #numEpochs = 100
    #multitask_model.fit(trainDataset, epochs=numEpochs,
    #                    validation_data=valDataset)
    ##############
    
    # save model
    #save_challenge_model(model_folder, multitask_model, trained_weights_file)
    multitask_model.save_weights(os.path.join(model_folder,trained_weights_file))
    
    if verbose >= 1:
        print('Done.')

     
#
# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):
    """ this function must return the keras model with its weights alredy loaded
    """
    filename = os.path.join(model_folder, trained_weights_file)
    multitask_model,  classification_model, cModel=\
        get_model(training=False, frameL=128, frameStepFactor=2)#, weights_file=filename)
    cModel.trainable = False
    multitask_model.load_weights(filename)
    return [multitask_model, classification_model]


# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def run_challenge_model(model, data, recordings, verbose):
    """ 
    model : output from load_challenge_model(...);
            keras classifier model: np-array -> 0/1 probabilities for murmur
    data:   output from load_patient_data(), helper_code,
            string with txt file content (some patient features like age,sex,..)
    recordings: output from load_recordings(), helper_code,
                list of np-arrays with the wav-files content
    """
    classification_model = model[1]
    
    recordings = [recording.astype(np.float32) for recording in recordings]
    numArrays = len(recordings)

    ## compute STFT for each wav
    
    # stft configurations SAME as in createMultitaskDataset
    frameL = 128
    frameStep = 64
    maxFreq = 65

    spectrums = []
    for array in recordings:
        spec = tf.signal.stft(array,frameL, frameStep)
        spectrums.append(np.abs(spec[:,:maxFreq]))

    ## put all specs in an 3-D array (padd with zeros)
    #
    maxL = np.max([spec.shape[0] for spec in spectrums])
    allspecs = np.zeros((len( spectrums),maxL, maxFreq),dtype=np.float32)
    for i,spec in enumerate(spectrums):
        auxLen = spec.shape[0]
        allspecs[i,:auxLen,:] = spec

    # apply model to all spectrums    
    logitsArray = classification_model.predict(allspecs)

    ## Murmur rules ###############################################

    #pdb.set_trace()
    # get maximum probabilities to have murmurs
    channelsClassificationMurmur = np.argmax(logitsArray[0],axis=-1)

    if np.any(channelsClassificationMurmur == 0):
        murmurLabel = 0
    elif  np.any(channelsClassificationMurmur == 1):
        murmurLabel = 1
    else:
        murmurLabel = 2

    ######### VERIFICAR E acrescentar os outcomes
    #probsArray = tf.math.softmax(logitsArray).numpy()
    probsMurmur = np.mean(logitsArray[0], axis=0)       
    classesMurmur = ['Present', 'Unknown', 'Absent']
    probsMurmur = probsMurmur[[1,2,0]]
    labelsMurmur = np.zeros(len(classesMurmur), dtype=np.int_)
    labelsMurmur[murmurLabel] = 1


    ## outcome rules

    # choose the best label for each channel
    channelsClassificationOutcome = logitsArray[1]
    
    # 'Abnormal' = 0
    outcomeNormalThreshold = 0.5 # above is normal
    outcomeNormalProbs = np.min(channelsClassificationOutcome)
    if outcomeNormalProbs <  outcomeNormalThreshold:
        outcomeLabel = 0 # Abnormal
    else:
        outcomeLabel = 1
    outcomeAbnormalProbs = 1 - outcomeNormalProbs
        
    outcomeLabels = np.zeros(2)
    outcomeLabels[outcomeLabel] = 1
    outcomeProbs = np.array([outcomeAbnormalProbs, outcomeNormalProbs])
    outcome_classes = ['Abnormal', 'Normal']

    classes = classesMurmur + outcome_classes
    labels = np.concatenate([labelsMurmur, outcomeLabels])
    probabilities = np.concatenate([probsMurmur, outcomeProbs])
    
    return classes, labels, probabilities



################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################

# Save your trained model.
def save_challenge_model(model_folder, model, weightsfname):
    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)
    filename = os.path.join(model_folder, weightsfname)
    model.save_weights(filename)

